@@ -31,17 +31,21 @@
 //! The best way to see how it is used is to read the `tests.rs` file;
 //! search for e.g. `UnitKey`.
 
-use std::marker;
 use std::fmt::Debug;
+use std::marker;
 use std::ops::Range;
 
+use snapshot_vec::{self as sv, UndoLog};
+use undo_log::{UndoLogs, VecLog};
+
 mod backing_vec;
-pub use self::backing_vec::{InPlace, UnificationStore};
+pub use self::backing_vec::{
+    Delegate, InPlace, UnificationStore, UnificationStoreBase, UnificationStoreMut,
+};
 
 #[cfg(feature = "persistent")]
 pub use self::backing_vec::Persistent;
 
-
 #[cfg(test)]
 mod tests;
 
@@ -65,13 +69,8 @@ pub trait UnifyKey: Copy + Clone + Debug + PartialEq {
 
     fn tag() -> &'static str;
 
-    /// If true, then `self` should be preferred as root to `other`.
-    /// Note that we assume a consistent partial ordering, so
-    /// returning true implies that `other.prefer_as_root_to(self)`
-    /// would return false.  If there is no ordering between two keys
-    /// (i.e., `a.prefer_as_root_to(b)` and `b.prefer_as_root_to(a)`
-    /// both return false) then the rank will be used to determine the
-    /// root in an optimal way.
+    /// You should return first the key that should be used as root,
+    /// then the other key (that will then point to the new root).
     ///
     /// NB. The only reason to implement this method is if you want to
     /// control what value is returned from `find()`. In general, it
@@ -156,10 +155,10 @@ pub struct NoError {
 /// time of the algorithm under control. For more information, see
 /// <http://en.wikipedia.org/wiki/Disjoint-set_data_structure>.
 #[derive(PartialEq, Clone, Debug)]
-pub struct VarValue<K: UnifyKey> { // FIXME pub
-    parent: K, // if equal to self, this is a root
+pub struct VarValue<K: UnifyKey> {
+    parent: K,       // if equal to self, this is a root
     value: K::Value, // value assigned (only relevant to root)
-    rank: u32, // max depth (only relevant to root)
+    rank: u32,       // max depth (only relevant to root)
 }
 
 /// Table of unification keys and their values. You must define a key type K
@@ -176,14 +175,21 @@ pub struct VarValue<K: UnifyKey> { // FIXME pub
 ///   - This implies that ordinary operations are quite a bit slower though.
 ///   - Requires the `persistent` feature be selected in your Cargo.toml file.
 #[derive(Clone, Debug, Default)]
-pub struct UnificationTable<S: UnificationStore> {
+pub struct UnificationTable<S: UnificationStoreBase> {
     /// Indicates the current value of each key.
     values: S,
 }
 
+pub type UnificationStorage<K> = Vec<VarValue<K>>;
+pub type UnificationTableStorage<K> = UnificationTable<InPlace<K, UnificationStorage<K>, ()>>;
+
 /// A unification table that uses an "in-place" vector.
 #[allow(type_alias_bounds)]
-pub type InPlaceUnificationTable<K: UnifyKey> = UnificationTable<InPlace<K>>;
+pub type InPlaceUnificationTable<
+    K: UnifyKey,
+    V: sv::VecLike<Delegate<K>> = Vec<VarValue<K>>,
+    L = VecLog<UndoLog<Delegate<K>>>,
+> = UnificationTable<InPlace<K, V, L>>;
 
 /// A unification table that uses a "persistent" vector.
 #[cfg(feature = "persistent")]
@@ -219,16 +225,25 @@ impl<K: UnifyKey> VarValue<K> {
         self.rank = rank;
         self.value = value;
     }
+}
 
-    fn parent(&self, self_key: K) -> Option<K> {
-        self.if_not_self(self.parent, self_key)
-    }
-
-    fn if_not_self(&self, key: K, self_key: K) -> Option<K> {
-        if key == self_key {
-            None
-        } else {
-            Some(key)
+impl<K> UnificationTableStorage<K>
+where
+    K: UnifyKey,
+{
+    /// Creates a `UnificationTable` using an external `undo_log`, allowing mutating methods to be
+    /// called if `L` does not implement `UndoLogs`
+    pub fn with_log<'a, L>(
+        &'a mut self,
+        undo_log: L,
+    ) -> UnificationTable<InPlace<K, &'a mut UnificationStorage<K>, L>>
+    where
+        L: UndoLogs<sv::UndoLog<Delegate<K>>>,
+    {
+        UnificationTable {
+            values: InPlace {
+                values: self.values.values.with_log(undo_log),
+            },
         }
     }
 }
@@ -238,11 +253,13 @@ impl<K: UnifyKey> VarValue<K> {
 // other type parameter U, and we have no way to say
 // Option<U>:LatticeValue.
 
-impl<S: UnificationStore> UnificationTable<S> {
+impl<S: UnificationStoreBase + Default> UnificationTable<S> {
     pub fn new() -> Self {
         Self::default()
     }
+}
 
+impl<S: UnificationStore> UnificationTable<S> {
     /// Starts a new snapshot. Each snapshot must be either
     /// rolled back or committed in a "LIFO" (stack) order.
     pub fn snapshot(&mut self) -> Snapshot<S> {
@@ -266,6 +283,27 @@ impl<S: UnificationStore> UnificationTable<S> {
         self.values.commit(snapshot.snapshot);
     }
 
+    /// Returns the keys of all variables created since the `snapshot`.
+    pub fn vars_since_snapshot(&self, snapshot: &Snapshot<S>) -> Range<S::Key> {
+        let range = self.values.values_since_snapshot(&snapshot.snapshot);
+        S::Key::from_index(range.start as u32)..S::Key::from_index(range.end as u32)
+    }
+}
+
+impl<S: UnificationStoreBase> UnificationTable<S> {
+    /// Returns the number of keys created so far.
+    pub fn len(&self) -> usize {
+        self.values.len()
+    }
+
+    /// Obtains the current value for a particular key.
+    /// Not for end-users; they can use `probe_value`.
+    fn value(&self, key: S::Key) -> &VarValue<S::Key> {
+        &self.values[key.index() as usize]
+    }
+}
+
+impl<S: UnificationStoreMut> UnificationTable<S> {
     /// Creates a fresh key with the given value.
     pub fn new_key(&mut self, value: S::Value) -> S::Key {
         let len = self.values.len();
@@ -284,10 +322,7 @@ impl<S: UnificationStore> UnificationTable<S> {
     /// Clears all unifications that have been performed, resetting to
     /// the initial state. The values of each variable are given by
     /// the closure.
-    pub fn reset_unifications(
-        &mut self,
-        mut value: impl FnMut(S::Key) -> S::Value,
-    ) {
+    pub fn reset_unifications(&mut self, mut value: impl FnMut(S::Key) -> S::Value) {
         self.values.reset_unifications(|i| {
             let key = UnifyKey::from_index(i as u32);
             let value = value(key);
@@ -295,26 +330,6 @@ impl<S: UnificationStore> UnificationTable<S> {
         });
     }
 
-    /// Returns the number of keys created so far.
-    pub fn len(&self) -> usize {
-        self.values.len()
-    }
-
-    /// Returns the keys of all variables created since the `snapshot`.
-    pub fn vars_since_snapshot(
-        &self,
-        snapshot: &Snapshot<S>,
-    ) -> Range<S::Key> {
-        let range = self.values.values_since_snapshot(&snapshot.snapshot);
-        S::Key::from_index(range.start as u32)..S::Key::from_index(range.end as u32)
-    }
-
-    /// Obtains the current value for a particular key.
-    /// Not for end-users; they can use `probe_value`.
-    fn value(&self, key: S::Key) -> &VarValue<S::Key> {
-        &self.values[key.index() as usize]
-    }
-
     /// Find the root node for `vid`. This uses the standard
     /// union-find algorithm with path compression:
     /// <http://en.wikipedia.org/wiki/Disjoint-set_data_structure>.
@@ -326,13 +341,12 @@ impl<S: UnificationStore> UnificationTable<S> {
     /// callsites. `uninlined_get_root_key` is the never-inlined version.
     #[inline(always)]
     fn inlined_get_root_key(&mut self, vid: S::Key) -> S::Key {
-        let redirect = {
-            match self.value(vid).parent(vid) {
-                None => return vid,
-                Some(redirect) => redirect,
-            }
-        };
+        let v = self.value(vid);
+        if v.parent == vid {
+            return vid;
+        }
 
+        let redirect = v.parent;
         let root_key: S::Key = self.uninlined_get_root_key(redirect);
         if root_key != redirect {
             // Path compression
@@ -370,13 +384,12 @@ impl<S: UnificationStore> UnificationTable<S> {
 
         let rank_a = self.value(key_a).rank;
         let rank_b = self.value(key_b).rank;
-        if let Some((new_root, redirected)) =
-            S::Key::order_roots(
-                key_a,
-                &self.value(key_a).value,
-                key_b,
-                &self.value(key_b).value,
-            ) {
+        if let Some((new_root, redirected)) = S::Key::order_roots(
+            key_a,
+            &self.value(key_a).value,
+            key_b,
+            &self.value(key_b).value,
+        ) {
             // compute the new rank for the new root that they chose;
             // this may not be the optimal choice.
             let new_rank = if new_root == key_a {
@@ -435,7 +448,29 @@ impl<S: UnificationStore> UnificationTable<S> {
 
 impl<S, K, V> UnificationTable<S>
 where
-    S: UnificationStore<Key = K, Value = V>,
+    S: UnificationStoreBase<Key = K, Value = V>,
+    K: UnifyKey<Value = V>,
+    V: UnifyValue,
+{
+    /// Obtains current value for key without any pointer chasing; may return `None` if key has been union'd.
+    #[inline]
+    pub fn try_probe_value<'a, K1>(&'a self, id: K1) -> Option<&'a V>
+        where
+            K1: Into<K>,
+            K: 'a,
+    {
+        let id = id.into();
+        let v = self.value(id);
+        if v.parent == id {
+            return Some(&v.value);
+        }
+        None
+    }
+}
+
+impl<S, K, V> UnificationTable<S>
+where
+    S: UnificationStoreMut<Key = K, Value = V>,
     K: UnifyKey<Value = V>,
     V: UnifyValue,
 {
@@ -537,7 +572,6 @@ where
     }
 }
 
-
 ///////////////////////////////////////////////////////////////////////////
 
 impl UnifyValue for () {
@@ -554,14 +588,11 @@ impl<V: UnifyValue> UnifyValue for Option<V> {
     fn unify_values(a: &Option<V>, b: &Option<V>) -> Result<Self, V::Error> {
         match (a, b) {
             (&None, &None) => Ok(None),
-            (&Some(ref v), &None) |
-            (&None, &Some(ref v)) => Ok(Some(v.clone())),
-            (&Some(ref a), &Some(ref b)) => {
-                match V::unify_values(a, b) {
-                    Ok(v) => Ok(Some(v)),
-                    Err(err) => Err(err),
-                }
-            }
+            (&Some(ref v), &None) | (&None, &Some(ref v)) => Ok(Some(v.clone())),
+            (&Some(ref a), &Some(ref b)) => match V::unify_values(a, b) {
+                Ok(v) => Ok(Some(v)),
+                Err(err) => Err(err),
+            },
         }
     }
 }
